@@ -1,6 +1,6 @@
 /*
     MythTV FUSE Module
-    Copyright (C) 2005-2008  Kees Cook <kees@outflux.net>
+    Copyright (C) 2005-2009  Kees Cook <kees@outflux.net>
 
     FUSE: Filesystem in Userspace
     Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>
@@ -14,12 +14,12 @@
     TODO:
         * switch to GString for all the string manipulations
         * add option for logging to syslog
-        * allow for configurable filename format
 
 */
 
 #include <config.h>
 
+#define _GNU_SOURCE
 #ifdef linux
 /* For pread()/pwrite() */
 #define _XOPEN_SOURCE 500
@@ -30,6 +30,9 @@
 // report on the calling paths
 #undef MYTHTVFS_DEBUG_FUNC
 
+// duplicated from programinfo.h
+#define FL_CUTLIST 0x0002
+
 #include <glib.h>
 #include <fuse.h>
 #include <fuse_opt.h>
@@ -49,6 +52,8 @@
 #include <strings.h> // rindex
 #include <time.h>    // strftime
 #include <string.h>  // memset
+#include <inttypes.h> // formatting
+#include <pthread.h> // thread locking
 
 // networking
 #include <sys/types.h>
@@ -63,9 +68,11 @@ struct show_t {
     char * description;
     char * callsign;
     char * pathname;
+    char * chanid;
     char * starttime;
     char * endtime;
     char * airdate;
+    int flags;
 };
 
 // options specific to mythtvfs
@@ -80,6 +87,9 @@ struct mythtvfs {
     char * base_path;
     char * host;
     int    port;
+    char * format;
+    char * date_format;
+    char * datetime_format;
     int short_display_flag;
     int backend_version_override; // manually override compile-time defaults
     int program_width_override;   // manually override compile-time defaults
@@ -88,10 +98,12 @@ struct mythtvfs {
     char * logfile_name;
 
     /* runtime state trackers */
+    pthread_mutex_t lock;
     size_t base_len;  // cached strlen(base_path)
     int backend;      // backend fd
     int program_cols; // how many columns are reported from this backend version
     FILE * logfile;
+    GHashTable *recordings; // hash table of recordings
 };
 
 static struct mythtvfs mythtvfs;
@@ -102,13 +114,16 @@ static struct fuse_opt mythtvfs_opts[] = {
     /* -o options */
     MYTHTVFS_OPT("host=%s",            host, 0),
     MYTHTVFS_OPT("port=%d",            port, 0),
+    MYTHTVFS_OPT("format=%s",          format, 0),
+    MYTHTVFS_OPT("date_format=%s",     date_format, 0),
+    MYTHTVFS_OPT("datetime_format=%s", datetime_format, 0),
     MYTHTVFS_OPT("short-display",      short_display_flag, 1),
     MYTHTVFS_OPT("logfile=%s",         logfile_name, 0),
     MYTHTVFS_OPT("backend-version=%d", backend_version_override, 0),
     MYTHTVFS_OPT("program-width=%d",   program_width_override, 0),
     MYTHTVFS_OPT("invalid-chars=%s",   invalid_chars, 0),
     MYTHTVFS_OPT("replacement-char=%s",replacement_char, 0),
-  
+
     FUSE_OPT_KEY("-V",                KEY_VERSION),
     FUSE_OPT_KEY("--version",         KEY_VERSION),
     FUSE_OPT_KEY("-h",                KEY_HELP),
@@ -129,6 +144,9 @@ static void usage(const char *progname)
         "MythTVfs options:\n"
         "    -o host=HOST          mythtv backend hostname\n"
         "    -o port=PORT          mythtv backend host port\n"
+        "    -o format=FORMAT      format for mythtvfs filenames\n"
+        "    -o date-format=FORMAT format for date components of filenames\n"
+        "    -o datetime-format=S  format for combined date and time components\n"
         "    -o short-display      shorter filename display\n"
         "    -o logfile=FILE       write verbose logging to FILE\n"
         "    -o backend-version=N  override compile-time backend version\n"
@@ -157,11 +175,11 @@ static char * aim_to_base(const char * oldpath)
     while (path && *path=='/') path++;
     /* make sure path is NULL for "empty" */
     if (path && *path=='\0') path=NULL;
-    
+
     /* calculate new path length */
     if (path) len = strlen(path); 
     len += mythtvfs.base_len + 1 /* "/" */ + 1 /* NULL */;
-    
+
     /* grow path length */
     if (aimed_len < len) {
         if (!(aimed_path = realloc(aimed_path,len))) {
@@ -183,12 +201,17 @@ static char * aim_to_base(const char * oldpath)
     return aimed_path;
 }
 
-static char * fixed_path = NULL;
-static size_t fixed_len  = 0;
+// hash table item
+struct mythtv_recording
+{
+    char * path;
+    gboolean stale;
+};
+
 static char * fixup(const char * oldpath)
 {
-    size_t len = 0;
     char * path = (char *)oldpath;
+    struct mythtv_recording * recording;
 
 #ifdef MYTHTVFS_DEBUG_AIM
     if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"fix '%s'\n",oldpath);
@@ -196,35 +219,15 @@ static char * fixup(const char * oldpath)
 
     /* make sure path is "" for "empty" */
     if (!path) path="";
-    
-    /* calculate new path length */
-    len = strlen(path) + 1 /* NULL */; 
+    while (*path == '/') { path++; }
 
-    if (fixed_len < len) {
-        if (!(fixed_path = realloc(fixed_path,len))) {
-            perror("realloc");
-            exit(1);
-        }
-        fixed_len = len;
+    recording = g_hash_table_lookup(mythtvfs.recordings, path);
+    if (recording) {
+        return aim_to_base(recording->path);
     }
-
-    /* trim metadata */
-    char * mark;
-    strncpy(fixed_path,path,len);
-    if (!(mark = rindex(fixed_path,'}'))) {
-        return aim_to_base(fixed_path);
+    else {
+        return aim_to_base(path);
     }
-    path = mark+1;
-    if (!(mark = rindex(path,'.'))) {
-        return aim_to_base(fixed_path);
-    }
-    *mark='\0';
-
-#ifdef MYTHTVFS_DEBUG_AIM
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"fixed '%s'\n",path);
-#endif
-
-    return aim_to_base(path);
 }
 
 //
@@ -255,7 +258,7 @@ int backend_init()
     }
     bzero((char*)&host_addr,sizeof(host_addr));
     host_addr.sin_family = AF_INET;
-    bcopy((char*)hostinfo->h_addr,
+    bcopy((char*)hostinfo->h_addr_list[0],
           (char*)&host_addr.sin_addr.s_addr,
           hostinfo->h_length);
     host_addr.sin_port = htons(mythtvfs.port);
@@ -400,7 +403,7 @@ void backend_version_check()
         mythtvfs.program_cols = versions[0].width;
         found = 1;
     }
-    
+
     // check for known versions
     for (idx = 0; !found && versions[idx].version; idx++) {
         if ( (version = _version_check(versions[idx].version))<0 ) continue;
@@ -438,6 +441,103 @@ void backend_done()
     backend_close();
 }
 
+struct cutlist_t {
+    int endmark;
+    long long location;
+    struct cutlist_t * next;
+};
+
+void free_cutlist(struct cutlist_t *cutlist)
+{
+    struct cutlist_t *next;
+
+    while (cutlist) {
+        next = cutlist->next;
+        free(cutlist);
+        cutlist = next;
+    }
+}
+
+// this function destroys the contents of *buffer (writes NULLs on separator)
+char * answer_chunk(char **buffer)
+{
+    if (!buffer) return NULL;
+    char * data = *buffer;
+    if (!data) return NULL;
+    char *sep;
+
+    if ((sep = strstr(data,MYTH_SEP))!=NULL) {
+        *sep = '\0';
+        sep+=strlen(MYTH_SEP);
+    }
+
+    *buffer = sep;
+    return data;
+}
+
+struct cutlist_t * cutlist_from_answer(char *answer)
+{
+    char * buffer = answer;
+    struct cutlist_t *cutlist_head = NULL;
+    struct cutlist_t *cutlist_tail = NULL;
+    struct cutlist_t *cutlist = NULL;
+    char * strval;
+    int items = 0;
+
+    if (!answer) return NULL;
+
+    strval = answer_chunk(&buffer);
+    items = atoi(strval);
+
+    while (items-- > 0) {
+        cutlist = calloc(1, sizeof *cutlist);
+        if (!cutlist) goto freak_out;
+        if (!cutlist_head) cutlist_head = cutlist;
+        if (!cutlist_tail) cutlist_tail = cutlist;
+        else {
+            cutlist_tail->next = cutlist;
+            cutlist_tail = cutlist_tail->next;
+        }
+
+        strval = answer_chunk(&buffer);
+        if (!strval) {
+            goto freak_out;
+        }
+        cutlist->endmark = atoi(strval);
+
+        strval = answer_chunk(&buffer);
+        if (!strval) {
+            goto freak_out;
+        }
+        cutlist->location = (long long)(atol(strval)) << 32;
+
+        strval = answer_chunk(&buffer);
+        if (!strval) {
+            goto freak_out;
+        }
+        cutlist->location |= (long long)(atol(strval)) & 0xffffffffLL;
+    }
+
+    free(answer);
+    return cutlist_head;
+
+freak_out:
+    free(answer);
+    free_cutlist(cutlist_head);
+    return NULL;
+}
+
+struct cutlist_t * backend_query_cutlist(char *chanid, char *starttime)
+{
+    char *answer = NULL, *request = NULL;
+    if (asprintf(&request, "QUERY_CUTLIST %s %s", chanid, starttime)<0 ||
+        !request) return NULL;
+    backend_cmd(request, &answer);
+    free(request);
+
+    return cutlist_from_answer(answer);
+}
+
 char * backend_query()
 {
     char * answer = NULL;
@@ -453,6 +553,7 @@ void clear_show(struct show_t * show)
     if (show->description) free(show->description);
     if (show->callsign) free(show->callsign);
     if (show->pathname) free(show->pathname);
+    if (show->chanid) free(show->chanid);
     if (show->starttime) free(show->starttime);
     if (show->endtime) free(show->endtime);
     if (show->airdate) free(show->airdate);
@@ -474,15 +575,16 @@ char * backend_pop_program(char * buffer, struct show_t * show)
         // make sure we're not null
         if (!data) data="";
 
+        /* Found in libs/libmythtv/programinfo.cpp ProgramInfo::ToStringList */
         switch (col) {
-            case 0: /* unknown */ break;
             case 1: show->title = strdup(data); break;
-                    //if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"\ttitle '%s'\n",show->title); break;
             case 2: show->subtitle = strdup(data); break;
             case 3: show->description = strdup(data); break;
             /*
             case 4: show->category = strdup(data); break;
+            */
             case 5: show->chanid = strdup(data); break;
+            /*
             case 6: show->channum = strdup(data); break;
             */
             case 7: show->callsign = strdup(data); break;
@@ -491,12 +593,15 @@ char * backend_pop_program(char * buffer, struct show_t * show)
             */
             case 9: show->pathname = strdup(data); break;
             /*
-            case 10: show->fs_high = strdup(data); break;
-            case 11: show->fs_low = strdup(data); break;
+            case 10: show->filesize_hi = strdup(data); break;
+            case 11: show->filesize_lo = strdup(data); break;
             */
             case 12: show->starttime = strdup(data); break;
             case 13: show->endtime = strdup(data); break;
             /*
+            case 14: show->duplicate = strdup(data); break;
+            case 15: show->shareable = strdup(data); break;
+            case 16: show->findid = strdup(data); break;
             case 17: show->hostname = strdup(data); break;
             case 18: show->sourceid = strdup(data); break;
             case 19: show->cardid = strdup(data); break;
@@ -510,16 +615,23 @@ char * backend_pop_program(char * buffer, struct show_t * show)
             case 27: show->recstartts = strdup(data); break;
             case 28: show->recendts = strdup(data); break;
             case 29: show->repeat = strdup(data); break;
-            case 30: show->flags = strdup(data); break;
+            */
+            case 30: show->flags = atoi(data); break;
+            /*
             case 31: show->recgroup = strdup(data); break;
             case 32: show->commfree = strdup(data); break;
             case 33: show->outputfilters = strdup(data); break;
             case 34: show->seriesid = strdup(data); break;
             case 35: show->programid = strdup(data); break;
             case 36: show->lastmodified = strdup(data); break;
-            case 37: show->recpriority = strdup(data); break;
+            case 37: show->stars = strdup(data); break;
             */
             case 38: show->airdate = strdup(data); break;
+            /*
+            case 39: show->hasairdate = strdup(data); break;
+            case 40: show->playground = strdup(data); break;
+            ...
+            */
             default:
                 break;
         }
@@ -528,77 +640,81 @@ char * backend_pop_program(char * buffer, struct show_t * show)
 
     return buffer;
 }
-    
 
-
-static int mythtvfs_getattr(const char *path, struct stat *stbuf)
+void mark_stale(gpointer key, gpointer value, gpointer user_data)
 {
-    int res;
-#ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
-#endif
-
-    res = lstat(fixup(path), stbuf);
-    if(res == -1)
-        return -errno;
-
-    return 0;
+    struct mythtv_recording * recording = (struct mythtv_recording*)value;
+    recording->stale = TRUE;
 }
 
-static int mythtvfs_readlink(const char *path, char *buf, size_t size)
+struct readdir_info
 {
-    int res;
-#ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
-#endif
-
-    res = readlink(fixup(path), buf, size - 1);
-    if(res == -1)
-        return -errno;
-
-    buf[res] = '\0';
-    return 0;
-}
-
-
-static int mythtvfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
-                       off_t offset, struct fuse_file_info *fi)
-{
-    (void) offset;
-    (void) fi;
-    (void) path;
+    fuse_fill_dir_t filler;
+    void * fuse_buf;
     struct stat st;
+};
 
-#ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
-#endif
+void fill_readdir_struct(gpointer key, gpointer value, gpointer user_data)
+{
+    char *name = (char *)key;
+    struct readdir_info * info = (struct readdir_info*)user_data;
 
-    memset(&st, 0, sizeof(st));
+    info->st.st_ino++;
+    info->st.st_mode = S_IFREG|S_IRUSR|S_IRGRP|S_IROTH;
+    info->filler(info->fuse_buf, name, &(info->st), 0);
+}
 
-    st.st_ino++;
-    st.st_mode = S_IFDIR|S_IRUSR|S_IRGRP|S_IROTH|S_IXUSR|S_IXGRP|S_IXOTH;
-    filler(buf, ".", &st, 0);
-    st.st_ino++;
-    filler(buf, "..", &st, 0);
+gboolean is_stale(gpointer key, gpointer value, gpointer user_data)
+{
+    struct mythtv_recording * recording = (struct mythtv_recording*)value;
+    return recording->stale;
+}
 
+void recording_free(gpointer ptr)
+{
+    struct mythtv_recording *recording = (struct mythtv_recording*)ptr;
+    if (!recording) return;
+    if (recording->path) free(recording->path);
+    free(recording);
+}
+
+void refresh_recordings()
+{
     struct show_t show;
     memset(&show, 0, sizeof(show));
 
     char * buffer = backend_query();
     char * data = buffer;
+    char smallbuf[2]; // handy single-char string for copying
+    smallbuf[0]=' '; smallbuf[1]='\0';
+
+    /* mark all hash items stale */
+    g_hash_table_foreach(mythtvfs.recordings, mark_stale, NULL);
+
     while ( (data = backend_pop_program(data,&show)) ) {
         char *metaname = NULL, *ptr = NULL;
 
-        /*
-        printf("Title   : %s\n",show.title);
-        printf("Subtitle: %s\n",show.subtitle);
-        printf("Callsign: %s\n",show.callsign);
-        printf("Pathname: %s\n",show.pathname);
-        printf("Start   : %s\n",show.starttime);
-        printf("End     : %s\n",show.endtime);
-        printf("Airdate : %s\n",show.airdate);
-        printf("\n");
-        */
+#ifdef MYTHTVFS_DEBUG_FUNC
+        if (mythtvfs.logfile) {
+            fprintf(mythtvfs.logfile,
+                "Title   : %s\n" \
+                "Subtitle: %s\n" \
+                "Callsign: %s\n" \
+                "Pathname: %s\n" \
+                "ChanId  : %s\n" \
+                "Start   : %s\n" \
+                "End     : %s\n" \
+                "Airdate : %s\n",
+                show.title,
+                show.subtitle,
+                show.callsign,
+                show.pathname,
+                show.chanid,
+                show.starttime,
+                show.endtime,
+                show.airdate);
+        }
+#endif
 
         char * filename;
         if (!(filename = rindex(show.pathname,'/'))) {
@@ -614,74 +730,86 @@ static int mythtvfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
             continue;
         }
 
-        if ( mythtvfs.short_display_flag ) {
-            // non-standard (SHORT) filename display
-            char duration[32];
-            char start[32];
-            char air[32];
-            time_t start_time = atoi(show.starttime);
-            time_t end_time = atoi(show.endtime);
-            time_t air_time = atoi(show.airdate);
-            snprintf(duration,32,"%ld",(end_time-start_time));
-            strftime(start,32,"%c",localtime(&start_time));
-            strftime(air,32,"%Y-%m-%d",localtime(&air_time));
-
-            int len = 2+strlen(show.title)+
-                2+strlen(air)+
-                2+strlen(show.subtitle)+
-                2+strlen(start)+
-                2+strlen(show.callsign)+
-                2+strlen(duration)+
-                strlen(filename)+
-                4+ /* ".mpg" */
-                1; /* NULL */
-            if (!(metaname=malloc(len))) {
-                perror("malloc");
-                exit(1);
+        if ((show.flags & FL_CUTLIST)==FL_CUTLIST && mythtvfs.logfile) {
+            fprintf(mythtvfs.logfile,"Path: '%s' ChanId: %s Starttime: %s\n",show.pathname,show.chanid,show.starttime);
+            struct cutlist_t *cutlist = backend_query_cutlist(show.chanid,show.starttime);
+            struct cutlist_t *walk;
+            for (walk=cutlist; walk; walk=walk->next) {
+                fprintf(mythtvfs.logfile,"%s: %lld\n",walk->endmark ? "end" : "start", walk->location);
             }
-            snprintf(metaname,len,"{%s}{%s}{%s}{%s}{%s}{%s}%s.mpg",
-                show.title,
-                air,
-                show.subtitle,
-                start,
-                show.callsign,
-                duration,
-                filename);
+            free_cutlist(cutlist);
+        }
+
+        char duration[32];
+        char start[64];
+        char air[64];
+        time_t start_time = atoi(show.starttime);
+        time_t end_time = atoi(show.endtime);
+        time_t air_time = atoi(show.airdate);
+        snprintf(duration,32,"%ld",(end_time-start_time));
+        strftime(start,64,mythtvfs.datetime_format,localtime(&start_time));
+        if (strchr(show.airdate,'-')==NULL) {
+            strftime(air,64,mythtvfs.date_format,localtime(&air_time));
         } else {
-            // standard (LONG) filename display
-            char duration[32];
-            char start[32];
-            char air[32];
-            time_t start_time = atoi(show.starttime);
-            time_t end_time = atoi(show.endtime);
-            time_t air_time = atoi(show.airdate);
-            snprintf(duration,32,"%ld",(end_time-start_time));
-            strftime(start,32,"%I.%M %p %a %b %d, %Y",localtime(&start_time));
-            strftime(air,32,"%Y-%m-%d",localtime(&air_time));
+            strncpy(air,show.airdate,64);
+        }
 
-            int len = 2+strlen(show.title)+
-                      2+strlen(air)+
-                      2+strlen(show.subtitle)+
-                      2+strlen(start)+
-                      2+strlen(show.callsign)+
-                      2+strlen(duration)+
-                      2+strlen(show.description)+
-                      strlen(filename)+
-                      4+ /* ".mpg" */
-                      1; /* NULL */
-            if (!(metaname=malloc(len))) {
+        // copy show info to metaname using format
+        int i; // index into format
+        int pos; // position in output; may be different than i
+        int len=2; // len of output so far; init with space for NULL and some spare
+        char *what=NULL; // what we are to copy
+        int formatlen = strlen(mythtvfs.format);
+
+        for (i=0, pos=0, len=2; i<formatlen; i++) {
+            if (mythtvfs.format[i]!='%') {
+                // copy literally
+                what=smallbuf;
+                what[0]=mythtvfs.format[i];
+            } else {
+                i++; // let's look at next character
+                if (i>=formatlen) {
+                    fprintf(stderr, "error: format ends with %%\n");
+                    exit(1);
+                }
+                switch (mythtvfs.format[i]) {
+                case 'T':
+                    what=show.title; break;
+                case 'a':
+                    what=air; break;
+                case 'S':
+                    what=show.subtitle; break;
+                case 's':
+                    what=start; break;
+                case 'c':
+                    what=show.callsign; break;
+                case 'd':
+                    what=duration; break;
+                case 'D':
+                    what=show.description; break;
+                case 'f':
+                    what=filename; break;
+                default:
+                    fprintf(stderr, "unknown format char %c\n", mythtvfs.format[i]);
+                 exit (1);
+                    break;
+                }
+            }
+
+            int whatlen=strlen(what);
+            len+=whatlen;
+
+            // realloc the space
+            if (!(metaname=realloc(metaname, len))) {
                 perror("malloc");
                 exit(1);
             }
-            snprintf(metaname,len,"{%s}{%s}{%s}{%s}{%s}{%s}{%s}%s.mpg",
-                     show.title,
-                     air,
-                     show.subtitle,
-                     start,
-                     show.callsign,
-                     duration,
-                     show.description,
-                     filename);
+
+            // copy what to current position
+            strncpy(metaname+pos, what, len-pos);
+
+            // update position
+            pos+=whatlen;
         }
 
         // filter out invalid characters
@@ -690,27 +818,109 @@ static int mythtvfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                 *ptr=*mythtvfs.replacement_char;
         }
 
-        st.st_ino++;
-        st.st_mode = S_IFREG|S_IRUSR|S_IRGRP|S_IROTH;
-
-        filler(buf, metaname, &st, 0);
-        free(metaname);
-
+        struct mythtv_recording *recording = g_hash_table_lookup(mythtvfs.recordings, metaname);
+        if (recording) {
+            free(metaname);
+        }
+        else {
+            recording = malloc(sizeof(*recording));
+            recording->path = strdup(filename);
+            g_hash_table_insert(mythtvfs.recordings, metaname, recording);
+        }
+        recording->stale = FALSE;
         clear_show(&show);
     }
     free(buffer);
 
+    /* remove anything that vanished from the recordings list */
+    g_hash_table_foreach_remove(mythtvfs.recordings, is_stale, NULL);
+}
+
+
+/*
+ * All the mythtvfs_* function are threaded entry points.  For now, lock
+ * all entry since the mythtvfs internals are not thread-safe yet.
+ */
+#define DO_LOCK if (pthread_mutex_lock(&(mythtvfs.lock))!=0) { perror("pthread_mutex_lock"); }
+#define DO_UNLOCK if (pthread_mutex_unlock(&(mythtvfs.lock))!=0) { perror("pthread_mutex_unlock"); }
+
+static int mythtvfs_getattr(const char *path, struct stat *stbuf)
+{
+    int res;
+#ifdef MYTHTVFS_DEBUG_FUNC
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
+#endif
+
+    DO_LOCK
+    res = lstat(fixup(path), stbuf);
+    DO_UNLOCK
+
+    if(res == -1)
+        return -errno;
+
     return 0;
 }
 
+static int mythtvfs_readlink(const char *path, char *buf, size_t size)
+{
+    int res;
+#ifdef MYTHTVFS_DEBUG_FUNC
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
+#endif
+
+    DO_LOCK
+    res = readlink(fixup(path), buf, size - 1);
+    DO_UNLOCK
+
+    if(res == -1)
+        return -errno;
+
+    buf[res] = '\0';
+    return 0;
+}
+
+static int mythtvfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
+                       off_t offset, struct fuse_file_info *fi)
+{
+    (void) offset;
+    (void) fi;
+    (void) path;
+    struct readdir_info info;
+
+#ifdef MYTHTVFS_DEBUG_FUNC
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
+#endif
+
+    memset(&(info.st), 0, sizeof(info.st));
+    info.fuse_buf = buf;
+    info.filler = filler;
+
+    info.st.st_ino++;
+    info.st.st_mode = S_IFDIR|S_IRUSR|S_IRGRP|S_IROTH|S_IXUSR|S_IXGRP|S_IXOTH;
+    info.filler(info.fuse_buf, ".", &(info.st), 0);
+    info.st.st_ino++;
+    info.filler(info.fuse_buf, "..", &(info.st), 0);
+
+    DO_LOCK
+    refresh_recordings();
+    g_hash_table_foreach(mythtvfs.recordings, fill_readdir_struct, &info);
+    DO_UNLOCK
+
+    return 0;
+}
+
+
 static int mythtvfs_mknod(const char *path, mode_t mode, dev_t rdev)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = mknod(fixup(path), mode, rdev);
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -721,10 +931,13 @@ static int mythtvfs_mkdir(const char *path, mode_t mode)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = mkdir(fixup(path), mode);
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -735,10 +948,13 @@ static int mythtvfs_unlink(const char *path)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = unlink(fixup(path));
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -749,10 +965,13 @@ static int mythtvfs_rmdir(const char *path)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = rmdir(fixup(path));
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -765,11 +984,13 @@ static int mythtvfs_symlink(const char *from, const char *to)
     char * fixed_from = NULL;
     char * fixed_to = NULL;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s %s\n",__FUNCTION__,from,to);
 #endif
 
+    DO_LOCK
     fixed_from = strdup(fixup(from));
     fixed_to   = strdup(fixup(to));
+    DO_UNLOCK
 
     res = symlink(fixed_from, fixed_to);
 
@@ -788,11 +1009,13 @@ static int mythtvfs_rename(const char *from, const char *to)
     char * fixed_from = NULL;
     char * fixed_to = NULL;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s %s\n",__FUNCTION__,from,to);
 #endif
 
+    DO_LOCK
     fixed_from = strdup(fixup(from));
     fixed_to   = strdup(fixup(to));
+    DO_UNLOCK
 
     res = rename(fixed_from, fixed_to);
 
@@ -811,11 +1034,13 @@ static int mythtvfs_link(const char *from, const char *to)
     char * fixed_from = NULL;
     char * fixed_to = NULL;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s %s\n",__FUNCTION__,from,to);
 #endif
 
+    DO_LOCK
     fixed_from = strdup(fixup(from));
     fixed_to   = strdup(fixup(to));
+    DO_UNLOCK
 
     res = link(from, to);
 
@@ -832,10 +1057,13 @@ static int mythtvfs_chmod(const char *path, mode_t mode)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = chmod(fixup(path), mode);
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -846,10 +1074,13 @@ static int mythtvfs_chown(const char *path, uid_t uid, gid_t gid)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = lchown(fixup(path), uid, gid);
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -860,10 +1091,13 @@ static int mythtvfs_truncate(const char *path, off_t size)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s @%" PRId64 "\n",__FUNCTION__,path,size);
 #endif
 
+    DO_LOCK
     res = truncate(fixup(path), size);
+    DO_UNLOCK
+
     if(res == -1)
         return -errno;
 
@@ -874,10 +1108,12 @@ static int mythtvfs_utime(const char *path, struct utimbuf *buf)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = utime(fixup(path), buf);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
 
@@ -889,10 +1125,12 @@ static int mythtvfs_open(const char *path, struct fuse_file_info *fi)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = open(fixup(path), fi->flags);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
 
@@ -906,11 +1144,24 @@ static int mythtvfs_read(const char *path, char *buf, size_t size, off_t offset,
     int fd;
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s @%" PRId64 " %" PRId64 "\n",__FUNCTION__,path,offset,size);
 #endif
 
     (void) fi;
+
+/*
+    // redirect .edl files
+    int len = strlen(path);
+    if (len>4 && strcmp(path+len-4,".edl")==0) {
+
+        ...
+        return res;
+    }
+*/
+
+    DO_LOCK
     fd = open(fixup(path), O_RDONLY);
+    DO_UNLOCK
     if(fd == -1)
         return -errno;
 
@@ -928,11 +1179,13 @@ static int mythtvfs_write(const char *path, const char *buf, size_t size,
     int fd;
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s @%" PRId64 " %" PRId64 "\n",__FUNCTION__,offset,size);
 #endif
 
     (void) fi;
+    DO_LOCK
     fd = open(fixup(path), O_WRONLY);
+    DO_UNLOCK
     if(fd == -1)
         return -errno;
 
@@ -954,10 +1207,12 @@ static int mythtvfs_statfs(const char *path, struct whichever_stat *stbuf)
 {
     int res;
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
+    DO_LOCK
     res = whichever_stat(fixup(path), stbuf);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
 
@@ -969,7 +1224,7 @@ static int mythtvfs_release(const char *path, struct fuse_file_info *fi)
     /* Just a stub.  This method is optional and can safely be left
        unimplemented */
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
     (void) path;
@@ -983,7 +1238,7 @@ static int mythtvfs_fsync(const char *path, int isdatasync,
     /* Just a stub.  This method is optional and can safely be left
        unimplemented */
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
 
     (void) path;
@@ -998,9 +1253,11 @@ static int mythtvfs_setxattr(const char *path, const char *name, const char *val
                         size_t size, int flags)
 {
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
+    DO_LOCK
     int res = lsetxattr(fixup(path), name, value, size, flags);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
     return 0;
@@ -1010,9 +1267,11 @@ static int mythtvfs_getxattr(const char *path, const char *name, char *value,
                     size_t size)
 {
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
+    DO_LOCK
     int res = lgetxattr(fixup(path), name, value, size);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
     return res;
@@ -1021,9 +1280,11 @@ static int mythtvfs_getxattr(const char *path, const char *name, char *value,
 static int mythtvfs_listxattr(const char *path, char *list, size_t size)
 {
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
+    DO_LOCK
     int res = llistxattr(fixup(path), list, size);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
     return res;
@@ -1032,9 +1293,11 @@ static int mythtvfs_listxattr(const char *path, char *list, size_t size)
 static int mythtvfs_removexattr(const char *path, const char *name)
 {
 #ifdef MYTHTVFS_DEBUG_FUNC
-    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s\n",__FUNCTION__);
+    if (mythtvfs.logfile) fprintf(mythtvfs.logfile,"%s %s\n",__FUNCTION__,path);
 #endif
+    DO_LOCK
     int res = lremovexattr(fixup(path), name);
+    DO_UNLOCK
     if(res == -1)
         return -errno;
     return 0;
@@ -1074,7 +1337,7 @@ static int mythtvfs_opt_proc(void *data, const char *arg, int key,
                              struct fuse_args *outargs)
 {
     (void) data;
-    
+
     switch (key) {
     case FUSE_OPT_KEY_OPT:
         return 1;
@@ -1085,7 +1348,7 @@ static int mythtvfs_opt_proc(void *data, const char *arg, int key,
             mythtvfs.base_len  = strlen(mythtvfs.base_path);
             return 0;
         }
-        
+
         return 1;
 
     case KEY_HELP:
@@ -1101,7 +1364,7 @@ static int mythtvfs_opt_proc(void *data, const char *arg, int key,
         fuse_main(outargs->argc, outargs->argv, &mythtvfs_oper, NULL);
 #endif
         exit(0);
-    
+
     default:
         fprintf(stderr, "internal error\n");
         abort();
@@ -1116,6 +1379,9 @@ int main(int argc, char *argv[])
     // setup "defaults"
     mythtvfs.base_path = NULL;
     mythtvfs.host      = NULL;
+    mythtvfs.format    = "{%T}{%a}{%S}{%s}{%c}{%d}{%D}%f.mpg";
+    mythtvfs.date_format    = "%Y-%m-%d";
+    mythtvfs.datetime_format    = "%I.%M %p %a %b %d, %Y";
     mythtvfs.port      = 6543;
     mythtvfs.short_display_flag = 0;
     mythtvfs.backend_version_override = 0;
@@ -1127,6 +1393,11 @@ int main(int argc, char *argv[])
     mythtvfs.logfile   = NULL;
     mythtvfs.logfile_name = NULL;
     mythtvfs.program_cols = 0;
+    mythtvfs.recordings = g_hash_table_new_full(g_str_hash, g_str_equal, free, recording_free);
+    if (pthread_mutex_init(&(mythtvfs.lock), NULL)!=0) {
+        perror("pthread_mutex_init");
+        exit(1);
+    }
 
     // process options
     if (fuse_opt_parse(&args,&mythtvfs,mythtvfs_opts,mythtvfs_opt_proc) == -1)
@@ -1142,6 +1413,11 @@ int main(int argc, char *argv[])
         exit(1);
     }
 
+    if (mythtvfs.short_display_flag) {
+        mythtvfs.format="{%T}{%a}{%S}{%s}{%c}{%d}%f.mpg";
+        mythtvfs.datetime_format = "%c";
+    }
+
     if (mythtvfs.logfile_name) {
         if (!(mythtvfs.logfile = fopen(mythtvfs.logfile_name,"a"))) {
             perror(mythtvfs.logfile_name);
